//-----------------------------------------------------------------------------------------
//------------------------------ SISTEMA DE DEFESA ----------------------------------------
//-----------------------------------------------------------------------------------------

#include <math.h>
#include<AccelStepper.h>
#include <SoftwareSerial.h>
#include <Stepper.h>
#include <IRremote.h>
#include <Servo.h>

// Inicia o Software Serial
SoftwareSerial Bluetooth(51, 53); // Arduino(RX, TX) - HC-05 Bluetooth (TX, RX)

// Define o receptor
const int RECV_PIN = 12;
IRrecv irrecv(RECV_PIN);
decode_results results;

// Define o servo
#define SERVO 11
Servo s;
int pos = 80; //Posição inicial definida pela localização do servo no mecanismo final

// Define o motor de passo do alimentador
const int passosAlimentador = 500; //Passos por revolução do motor
Stepper alimentador(passosAlimentador, 31, 35, 33, 37);  //Ligação do motor utilizando as portas de 22 a 26

// Define os pinos do relé
int relay1 = 22;
int relay2 = 24;
int relay3 = 26;
int relay4 = 28;

//Define os motores de passo e os pinos que eles utilizarão
//(driver, STEP, DIR)
AccelStepper m1(1, 3, 2);  // Stepper1 = Esquerda frontal
AccelStepper m2(1, 5, 4); // Stepper2 = Direita frontal
AccelStepper m4(1, 44, 42); //Stepper 3 = Esquerda traseira
AccelStepper m3(1, 46, 48); // Stepper 4 = Direita traseira

//Array para armazenar todas as informações recebidas. 255, x, y, z, est, ang, fire.
int dataIn[8] {0, 0, 0, 0, 0, 0, 0};
int in_byte = 0;
int array_index = 0;

//Variáveis de armazenamento
int luz = 0;
int relay = 0;
int ang;
int inc1;
int inc2;
int est;
int fire;
int fire_2 = 0;
int vx;
int vy;
int w;

//Variáveis da velocidade
float v1;
float v2;
float v3;
float v4;
int V1;
int V2;
int V3;
int V4;

//Variáveis para iteração
int i = 0;
int x = 0;
int c1 = 0;
int c2 = 0;
int alvo = 0;

//Variáveis dos timers, com millis()
unsigned long antes_bt = 0;
unsigned long antes_servo1 = 0;
unsigned long antes_servo2 = 0;
unsigned long antes_stepper1 = 0;
unsigned long antes_stepper2 = 0;
unsigned long antes_disparo = 0;
unsigned long agora;
int intervalo = 100;


//-----------------------------------------------------------------------------------------


void setup() {

  //Inicia as comunicações seriais
  Serial.begin(9600);
  Bluetooth.begin(9600);

  //Seta as velocidades máximas dos motores de passo
  m1.setMaxSpeed(4000);
  m2.setMaxSpeed(4000);
  m3.setMaxSpeed(4000);
  m4.setMaxSpeed(4000);

  //Seta as aceleraçoes máximas dos motores de passo
  m1.setAcceleration(5000);
  m2.setAcceleration(5000);
  m3.setAcceleration(5000);
  m4.setAcceleration(5000);

  //Seta a velocidade do alimentador
  alimentador.setSpeed(60);

  //Seta os pinos do relé como saída
  pinMode(relay1, OUTPUT);
  pinMode(relay2, OUTPUT);
  pinMode(relay3, OUTPUT);
  pinMode(relay4, OUTPUT);

  //Inicia o sistema com as saídas do relé desligadas
  digitalWrite(relay1, HIGH);
  digitalWrite(relay2, HIGH);
  digitalWrite(relay3, HIGH);
  digitalWrite(relay4, HIGH);

  //Conecta o servo do pino 11 no objeto servo e inicia ele em uma posição prédefinida
  s.attach(SERVO);
  s.write(pos);

  //Ativa o receptor infravermelho do pino 12
  pinMode(RECV_PIN, INPUT);
  irrecv.enableIRIn();
  irrecv.blink13(true);
}

//-----------------------------------------------------------------------------------------


void loop() {

  //Inicia o loop com a recepção de sinais do bluetooth
  dados();


  //Executa os diferentes modos de operação conforme o dado recebido
  switch (est) {
    case 1:
      manual();
      break;

    case 2:
      automatico();
      break;
  }
}

//-----------------------------------------------------------------------------------------

void dados() {

  if (Bluetooth.available() > 0) {  //Se o bluetooth estiver recebendo algum dado executa a função
    in_byte = Bluetooth.read();    //Armazena o byte na variável de entrada
    if (in_byte == (255)) {       //Se a variável é 0 seta o índice do array para 0. Isso garante que cada número ocupe a posição correta.
      array_index = 0;
    }
    dataIn[array_index] = in_byte;   //Armazena o byte recebido no array com o índice referente
    array_index = array_index + 1;  //Itera o índice
  }

  //Endereça os itens do array para a variável correspondente
  vx = (float(dataIn[1]) - 100);
  vy = (100 - float(dataIn[2]));
  w =  (float(dataIn[3]) - 100);
  est = dataIn[4];
  inc1 = dataIn[5];
  fire = dataIn[6];
  luz = dataIn[7];
  relay = dataIn[8];
}

//-----------------------------------------------------------------------------------------

void velocidades() {

  //Calcula as velocidades conforme as equações de movimento, utilizando um fator para o ajuste de amplitude
  v1 = (-vy - vx - w) * (4);
  v2 = ( vy - vx - w) * (4);
  v3 = (-vy + vx - w) * (4);
  v4 = ( vy + vx - w) * (4);

  //Arredonda as velocidades para serem utilizadas como argumento nas funções da biblioteca AccelStepper
  V1 = round(v1);
  V2 = round(v2);
  V3 = round(v3);
  V4 = round(v4);

  //Define as velocidades atuais dos motores
  m1.setSpeed(V1);
  m2.setSpeed(V2);
  m3.setSpeed(V3);
  m4.setSpeed(V4);

  //Executa as velocidades
  m1.runSpeed();
  m2.runSpeed();
  m3.runSpeed();
  m4.runSpeed();
}

//-----------------------------------------------------------------------------------------


void atirar_manual() {

  //Inicia com as variáveis do disparo zeradas e os motores desligados
  if (fire == 0) {
    fire_2 = 0;
    digitalWrite(relay2, HIGH);
  }

  //O primeior incremento do disparo aciona os motores
  if (fire != fire_2) {
    fire_2 = fire;
    if (fire == 1) {
      digitalWrite(relay2, LOW);
    }

    //Após o acionamento cada incremento alimenta uma bolinha, rotacionando o stepper 40 graus
    else {
      alimentador.step(-227);
    }
  }
}

//-----------------------------------------------------------------------------------------

void atirar_automatico() {

  //Inicia utilizando um timer com millis() de intervalo variável
  agora = millis();
  if (agora - antes_disparo >= intervalo) {
    antes_disparo = agora;
    if (c2 == 0) {   //Na primeira execução do timer os motores são acionados e o próximo intervalo definido como 10 segundos
      digitalWrite(relay2, LOW);
      intervalo = 10000;
      c2 = 1;
    }
    else if ( c2 < 4) {   //Nas próximas 3 execuções do timer, com intervalos de 2 segundos, bolinhas são alimentadas
      alimentador.step(-227);
      intervalo = 2000;
      c2 = c2 + 1;
    }
    else if (c2 == 4) {  //Na quarta execução do timer os motores são desativados e as variáveis do alvo reiniciadas
      digitalWrite(relay2, HIGH);
      alvo =  0;
      c1 = 0;
      s.write(80);
    }
  }
}

//-----------------------------------------------------------------------------------------

void manual() {

  //Reseta as variáveis e desliga os motores;
  if (alvo == 1) {
    digitalWrite(relay2, HIGH);
  }
  alvo = 0;
  c1 = 0;
  c2 = 0;

  //Atualiza as velocidades dos motores
  velocidades();

  //Atualiza a posição do servo
  servo();

  //Atualiza as saídas do relé
  rele();

  //Executa a rotina de disparo
  atirar_manual();
}

//-----------------------------------------------------------------------------------------

void automatico() {

  //Reseta as variáveis e desliga os motores;
  if (fire_2 != 0) {
    fire_2 = 0;
    digitalWrite(relay2, HIGH);
  }

  //Enquanto o alvo não for achado, executa a rotina de checagem
  if ((alvo == 0) && (est == 2)) {
    rotacionar();                     //Rotaciona os motores
    enviar();                        //Envia os dados atuais por bluetiith
    if (irrecv.decode(&results)) {  //Executa caso o receptor detecte o alvo
      irrecv.resume();             //Desliga o receptor
      s.write(i-30);              //Compensa a inclinação do disparador
      alvo = 1;                  //Atualiza as variáveis do alvo
      inc2 = i;
      i = 10;
      enviar2();              //Envia os dados atuais por bluetooth
    }

    //Após 500 iterações executa a varredura
    if ((c1 == 500) && (alvo == 0)) {
      parar();                        //Pausa a rotação dos motores
      for (i = 90; i >= 10; i--) {   //Movimenta o servo e checa o sensor
        s.write(i);
        delay(100);
        enviar();
        if (irrecv.decode(&results)) {
          irrecv.resume();
          s.write(i-30);
          alvo = 1;
          inc2 = i;
          i = 10;
          enviar2();
        }
      }
      //Se o alvo não for achado retorna o servo à posição inicial e reincia a contagem
      if ( alvo == 0) {
        s.write(90);
        delay(100);
        c1 = 0; //Reinicia a contagem
      }
      i = 90;
    }
    //Itera as variáveis de posição
    c1 = c1 + 1;
  }
  //Se o alvo for encontrado executa o disparo
  if ((alvo == 1) && (est == 2))  {
    parar();
    atirar_automatico();
    enviar2();
  }
}

//-----------------------------------------------------------------------------------------

void servo() {

  //Utilizando a função millis(), a cada 100ms, atualiza a posição do servo
  agora = millis();
  if (agora - antes_servo1 >= 100) {
    antes_servo1 = agora;
    s.write(100 - inc1);
  }
}

//-----------------------------------------------------------------------------------------

void rotacionar() {

  //Move o servo para a posição inicial
  s.write(80);

  //Seta as velocidades para uma rotação precisa
  m1.setSpeed(200);
  m2.setSpeed(200);
  m3.setSpeed(200);
  m4.setSpeed(200);

  //Executa as velocidades
  m1.runSpeed();
  m2.runSpeed();
  m3.runSpeed();
  m4.runSpeed();
}

//-----------------------------------------------------------------------------------------

void parar() {

  //Pausa a rotação dos motores
  m1.stop();
  m2.stop();
  m3.stop();
  m4.stop();
}

//-----------------------------------------------------------------------------------------

void enviar() {

    //Envia os dados do sistema via bluetooth, utilizando a formatação e os espaçadores corretos
    Bluetooth.print("d");
    Bluetooth.print(c1);
    Bluetooth.print("|");
    Bluetooth.print(i);
    Bluetooth.print("|");
    Bluetooth.print(alvo);
}

void enviar2() {

    //Envia os dados do sistema, pós detecção, via bluetooth, utilizando a formatação e os espaçadores corretos
    Bluetooth.print("d");
    Bluetooth.print(c1);
    Bluetooth.print("|");
    Bluetooth.print(inc2);
    Bluetooth.print("|");
    Bluetooth.print(alvo);
}

//-----------------------------------------------------------------------------------------

void rele() {

  //Aciona o pino do relé correspondente às luzes conforme a variável recebida
  if (luz == 1) {
    digitalWrite(relay3, LOW);
  }
  else {
    digitalWrite(relay3, HIGH);
  }
}

//-----------------------------------------------------------------------------------------
